import os
import numpy as np
from scipy.constants import pi
import h5py


def read_scalar(path, keyword):
    scalars = os.path.join(path, 'scalars.txt')

    data = []
    with open(scalars, mode='r') as f:
        data_start = False
        data_loc = None
        for line in f.readlines():
            if line.find(keyword) > 0:
                if line.strip('#').split()[0].isnumeric():
                    continue
                data_start = True
                data_loc = line.strip('#').split().index(keyword)
                continue
            if data_start:
                data.append(float(line.split()[data_loc]))
    return np.asarray(data, dtype=float)

def get_timesteps(result_path, number=0):
    with h5py.File(os.path.join(result_path, f'Fields{number}.h5'), 'r') as h5f:
        ts = list(h5f['data'].keys())   
    return ts


def get_extent(result_path, number=0, component='Ey', lambda0=0.8e-6):
    '''
    get 2D extent
    '''
    with h5py.File(os.path.join(result_path, f'Fields{number}.h5'), 'r') as h5f:
        ts = list(h5f['data'].keys())
        nx, ny = h5f['data'][ts[0]][component].shape
        dx, dy = h5f['data'][ts[0]][component].attrs['gridSpacing']

    return np.array([0, nx*dx, 0, ny*dy]) / 2/pi * lambda0 /1e-6


def get_cellsize(result_path, number=0, component='Ey', lambda0=0.8e-6):
    '''
    get 2D or 3D cell size in um
    '''
    with h5py.File(os.path.join(result_path, f'Fields{number}.h5'), 'r') as h5f:
        ts = list(h5f['data'].keys())
        dset = h5f['data'][ts[0]][component]
        if dset.dims == 2:
            dx, dy = dset.attrs['gridSpacing']
            return dx/2/pi*lambda0/1e-6, dy/2/pi*lambda0/1e-6

        if dset.dims == 3:
            dx, dy, dz = dset.attrs['gridSpacing']
            return dx/2/pi*lambda0/1e-6, dy/2/pi*lambda0/1e-6, dz/2/pi*lambda0/1e-6

def get_field(result_path, ts, component, number=0, slice=()) -> np.ndarray:
    if isinstance(ts, int):
        ts = f'{ts:010d}'
    with h5py.File(os.path.join(result_path, f'Fields{number}.h5'), 'r') as h5f:
        dset = h5f['data'][ts][component]
        if len(dset.shape) == 2:
            return dset[slice].T
        else:
            return dset[slice]


def get_traj(result_path, name, component):
    with h5py.File(os.path.join(result_path, f'TrackParticles_{name}.h5'), 'r') as h5f:
        return h5f[component][()]


import os
import re
import sdf
import h5py
import numpy as np
from scipy.constants import m_e, c

_field_names = {
        'Ex': 'Electric Field/Ex', 'Ey' : 'Electric Field/Ey' , 'Ez' : 'Electric Field/Ez' ,
        'Bx' : 'Magnetic Field/Bx', 'By' : 'Magnetic Field/By' , 'Bz' : 'Magnetic Field/Bz' ,
}


def get_sdffiles(path, prefix=''):
    sdffiles = []
    with open(os.path.join(path, f'{prefix}.visit'), 'r') as f:
        sdffiles = f.read().splitlines()
    
    for i in range(len(sdffiles)):
        sdffiles[i] = os.path.join(path, sdffiles[i])

    return sdffiles


def get_extent(result_path, ts=None, prefix=''):
    '''
    get 2D extent
    '''
    if result_path[-4:] == '.sdf':
        sdffile_name = result_path
    elif isinstance(ts, int):
        ts = f'{ts:04d}'
        sdffile_name = f'{result_path}/{prefix}{ts}.sdf'
    f = sdf.read(sdffile_name, dict=True)
    
    extent = [
        f['Grid/Grid'].data[0][0],
        f['Grid/Grid'].data[0][-1],
        f['Grid/Grid'].data[1][0],
        f['Grid/Grid'].data[1][-1],
    ]
    
    return np.array(extent)


def get_field(result_path, component, ts=None, prefix='', slice=()) -> np.ndarray:
    if result_path[-4:] == '.sdf':
        sdffile_name = result_path
    elif isinstance(ts, int):
        ts = f'{ts:04d}'
        sdffile_name = f'{result_path}/{prefix}{ts}.sdf'
    
    f = sdf.read(sdffile_name, dict=True)
    
    if component in _field_names.keys():
        dset = f.pop(_field_names[component]).data
    else:
        dset = f.pop(f'Derived/Number_Density/{component}').data
        
    if len(dset.shape) == 2:
        return dset[slice].T
    else:
        return dset[slice]


def get_particles(result_path, name, component : str, ts=None, prefix='') -> np.ndarray:
    if result_path[-4:] == '.sdf':
        sdffile_name = result_path
    elif isinstance(ts, int):
        ts = f'{ts:04d}'
        sdffile_name = f'{result_path}/{prefix}{ts}.sdf'

    f = sdf.read(sdffile_name, dict=True)
    
    if component not in ['px', 'py', 'pz', 'x', 'y', 'z', 'w', 'id']:
        raise KeyError(f'{component} not supported.')

    if component in ['px', 'py', 'pz']:
        dset = f.pop(f'Particles/{component.capitalize()}/{name}').data
        return dset[()] / m_e / c
    
    if component in ['x', 'y', 'z']:
        dset = f.pop(f'Grid/Particles/{name}').data
        return dset[{'x' : 0, 'y' : 1, 'z' : 2}[component]]

    if component == 'w':
        dset = f.pop(f'Particles/Weight/{name}').data
        return dset[()]

    if component == 'id':
        dset = f.pop(f'Particles/ID/{name}').data
        return dset[()]
    

def _get_all_ids(sdffiles, name, comm):
    nt = len(sdffiles)
    rank = 0 if comm is None else comm.Get_rank()
    comm_size = 1 if comm is None else comm.Get_size()

    ids = set()
    for i in range(rank, nt, comm_size):
        sdffile = sdffiles[i]

        f = sdf.read(sdffile,dict=True)

        ids.update(set(f[f'Particles/ID/{name}'].data[()])) 
    
    if comm is not None:
        ids_all = set()
        ids = comm.Gater(ids, root=0)
        
        for ids_ in ids:
            ids_all.update(ids_)
        return np.array(ids_all, dtype=int)
    else:
        return np.array(ids, dtype=int)
        
        
        
def sort(result_path, name, selected_ids=None, props=None, prefix='', comm=None):
    all_props = {
        'vx' : f'Particles/Vx/{name}',
        'vy' : f'Particles/Vy/{name}',
        'vz' : f'Particles/Vz/{name}',
        'px' : f'Particles/Px/{name}',
        'py' : f'Particles/Py/{name}',
        'pz' : f'Particles/Pz/{name}',
        'x'  : f'Grid/Particles/{name}', 
        'y'  : f'Grid/Particles/{name}', 
        # 'z'  : f'Grid/Particles/{name}', 
    }
    sdffiles = get_sdffiles(result_path, prefix)
    if props is None:
        f = sdf.read(sdffiles[-1], dict=True)
        props = []
        for prop in all_props.keys():
            if all_props[prop] in f.keys():
                props.append(prop)
    else:
        pass

    nt = len(sdffiles)
    rank = 0 if comm is None else comm.Get_rank()
    comm_size = 1 if comm is None else comm.Get_size()

    if selected_ids is None:
        selected_ids = _get_all_ids(sdffiles, name, comm)
    else:
        selected_ids = set(selected_ids)

    npart = len(selected_ids)
    print(f'{npart} particles selected')

    ordermap = {id_ : i for i, id_ in enumerate(selected_ids)}

    mpio_kws = {}
    if comm is not None:
        mpio_kws = {'driver' : 'mpio', 'comm' : comm}
    with h5py.File(os.path.join(result_path, f'TrackParticles_{prefix}_{name.split("/")[-1]}.h5'), 'w', **mpio_kws) as h5f:
        for prop in props:
            h5f.create_dataset(prop, (nt, npart), fillvalue=np.nan, dtype='float64')
        h5f.create_dataset('t', (nt, ), fillvalue=np.nan, dtype='float64')
        
        for i in range(rank, nt, comm_size):
            sdffile = sdffiles[i]
            print(sdffile)
            f = sdf.read(sdffile,dict=True)
            ids = f[f'Particles/ID/{name}'].data[()]

            if len(set(ids) & selected_ids) == 0:
                continue

            # selected = np.any(selected_ids[:, np.newaxis] == ids, axis=0)
            selected = np.array([id in selected_ids for id in ids])
            ids = ids[selected]
            orders = [ordermap[id] for id in ids]
            
            for prop in props:
                buf = np.full(npart, np.nan)
                if prop in ['x', 'y', 'z']:
                    buf[orders] = f[all_props[prop]].data[{'x':0, 'y':1, 'z':2}[prop]][selected]
                else:
                    buf[orders] = f[all_props[prop]].data[()][selected]
                h5f[prop][i, :] = buf
            h5f['t'][i] = f['Header']['time']



from typing import Tuple
from numpy import sqrt, arctan, sin, cos, exp
from scipy.constants import mu_0, c, e, m_e, pi


def a0_from_E0(E0: float, lambda0: float = 0.8e-6) -> float:
    omega0 = 2*pi*c/lambda0
    return E0 * e / m_e / c / omega0


def E0_from_a0(a0: float, lambda0: float = 0.8e-6) -> float:
    omega0 = 2*pi*c/lambda0
    return a0 * m_e * c * omega0 / e


def I_from_E0(E0: float) -> float:
    return 0.5 * E0**2 / mu_0 / c


def E0_from_I(I: float) -> float:
    return sqrt(I*mu_0*c*2)


def I_from_a0(a0: float, lambda0: float = 0.8e-6) -> float:
    E0 = E0_from_a0(a0, lambda0)
    return I_from_E0(E0)


def a0_from_I(I: float, lambda0: float = 0.8e-6) -> float:
    E0 = E0_from_I(I)
    return a0_from_E0(E0, lambda0)


def power(a0: float, w0: float, lambda0: float = 0.8e-6) -> float:
    I = I_from_a0(a0, lambda0)
    return 0.5 * pi * w0**2 * I


def E0_from_power(power: float, w0: float, lambda0: float = 0.8e-6) -> float:
    I = 2*power / (pi*w0**2)
    return E0_from_I(I)


def a0_from_power(power: float, w0: float, lambda0: float = 0.8e-6) -> float:
    E0 = E0_from_power(power, w0, lambda0)
    return a0_from_E0(E0, lambda0)


def w0_from_power(power: float, a0: float, lambda0: float = 0.8e-6) -> float:
    I = I_from_a0(a0, lambda0)
    return sqrt(power/I/(0.5*pi))


def GaussianSpaceProfile(
    amp: float, w0: float, ctau: float, x0: Tuple[float, float, float],
    xf: float = None, lambda0: float = 0.8e-6, cep: float = 0.0
):
    k0 = 2*pi/lambda0
    zR = pi*w0**2/lambda0

    if xf is None:
        xf = x0[0]

    def profile2d(x, y):
        x -= xf
        y -= x0[1]

        r2 = y**2
        wz = w0 * sqrt(1 + (x/zR)**2)
        Rz = x * (1 + (zR/x)**2)
        gouy = arctan(x/zR)

        phi = k0 * (x + xf - x0[0]) + k0*r2/2/Rz - gouy
        return amp * cos(phi+cep) * w0/wz * exp(-r2/wz**2) * exp(-phi**2/(k0*ctau)**2)

    def profile3d(x, y, z):
        x -= xf
        y -= x0[1]
        z -= x0[2]

        r2 = y**2 + z**2
        wz = w0 * sqrt(1 + (x/zR)**2)
        Rz = x * (1 + (zR/x)**2)
        gouy = arctan(x/zR)

        phi = k0 * (x + xf - x0[0]) + k0*r2/2/Rz - gouy
        return amp * cos(phi+cep) * w0/wz * exp(-r2/wz**2) * exp(-phi**2/(k0*ctau)**2)

    if len(x0) == 2:
        return profile2d
    if len(x0) == 3:
        return profile3d



from scipy.constants import c, e, m_e, epsilon_0, pi, hbar
from numpy import sqrt

def nc(lambda0 : float) -> float:
    omega0 = 2*pi*c/lambda0
    return epsilon_0 * m_e * omega0**2 / e**2

def Pc(ne : float, nc : float) -> float:
    return 17 * (nc/ne) * 1e9 # W

def matched_n0(lambda0 : float, a0 : float, w0 : float) -> float:
    nc_ = nc(lambda0)
    return 1/pi**2 * a0 * lambda0**2 / w0**2 * nc_

def chi(Ex=0, Ey=0, Ez=0, Bx=0, By=0, Bz=0, ux=0, uy=0, uz=0):
    factor = e*hbar / (m_e**2 * c**3)
    
    gamma = sqrt(ux**2 + uy**2 + uz**2 + 1)

    chi = factor * sqrt(
        (gamma*Ex + (uy*Bz - uz*By)*c)**2 +
        (gamma*Ey + (uz*Bx - ux*Bz)*c)**2 +
        (gamma*Ez + (ux*By - uy*Bx)*c)**2 -
        (ux*Ex + uy*Ey + uz*Ez)**2
    )
    return chi


from numpy import concatenate, flipud, exp, pi, s_
import os 
import re

def get_h5files(path):
    h5files = []
    for file in os.listdir(path):
        if re.match(r'data\d*\.h5', file):
            h5files.append(f'{path}/{file}')

    h5files.sort()
    return h5files

def get_field_slice(field, theta=0, inversed=False):
    field_up = get_field(field, theta)
    field_down = get_field(field, theta+pi)
    fac = -1 if inversed else 1
    return concatenate((fac*flipud(field_down), field_up), axis=0)

def get_field(field, theta=0):
    modes = (field.shape[0] + 1) // 2
    data = field[0]
    for m in range(1, modes):
        data += ((field[2*m-1] + 1j*field[2*m]) * exp(-1j*m*theta)).real
        
    return(data)

def get_field_partial(field, theta=0, sr=s_[:], sz=s_[:]):
    modes = (field.shape[0] + 1) // 2
    data = field[0, sr, sz]
    for m in range(1, modes):
        data += ((field[2*m-1, sr, sz] + 1j*field[2*m, sr, sz]) * exp(-1j*m*theta)).real

    return(data)